"""
This class takes a diff and creates query executors for each diff type.
It also makes the split between using SQL, ORM or any other way to execute the queries.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from typing import Optional

from .abstract_orm import AbstractORM
from .orm_creator import InsertOrmCreator, UpdateOrmCreator, DeleteOrmCreator
from .sql_creator import InsertSqlCreator, UpdateSqlCreator, DeleteSqlCreator
from ..stml.alias_enricher import AliasEnricher


class DiffToExecutor:

    def __init__(self):
        pass

    def diff_executor(self, mapping, diffs, context=None, orm: Optional[AbstractORM] = None):

        if not self._use_orm(mapping):
            # create SQL query executors
            return self._sql_executor(mapping, diffs, context)
        else:
            # assert that orm exists
            assert orm is not None, 'ORM is required for this mapping'

            # create ORM executors
            return self._orm_executor(mapping, diffs, context, orm)

    def _use_orm(self, mapping):
        # hard coded for now
        return mapping.name in ['ir_attachment']

    def _sql_executor(self, mapping, diffs, context=None):
        # get from tuple
        inserts, updates, deletes = diffs

        # add alias and parameter names to mapping before creating sql
        aliased_mapping = AliasEnricher().enrich(mapping)

        # create sql for each diff
        insert_sql = list(InsertSqlCreator().create_executors(aliased_mapping, inserts, context))
        update_sql = list(UpdateSqlCreator().create_executors(aliased_mapping, updates, context))
        delete_sql = list(DeleteSqlCreator().create_executors(aliased_mapping, deletes, context))

        return insert_sql + update_sql + delete_sql

    def _orm_executor(self, mapping, diffs, context, orm):

        # get from tuple
        inserts, updates, deletes = diffs

        # add alias and parameter names to mapping before creating sql
        aliased_mapping = AliasEnricher().enrich(mapping)

        # create sql for each diff
        insert_orm = list(InsertOrmCreator().create_executors(aliased_mapping, inserts, context, orm))
        update_orm = list(UpdateOrmCreator().create_executors(aliased_mapping, updates, context, orm))
        delete_orm = list(DeleteOrmCreator().create_executors(aliased_mapping, deletes, context, orm))

        return insert_orm + update_orm + delete_orm



'''
This class compiles a mapping into a where clause that does not recurse into foreign keys. Used for update and delete statements.

There's subtle difference between update/delete statements and the select statement that we need to find the primary key for ORM queries:
- update: update books set ... where ... book.authorid = authors.author_id
- delete: delete from books using authors where ... book.authorid = authors.author_id
- primary key selector: select ... from books join authors on book.authorid = authors.author_id

Author: Romke Jonker
Email: romke@stml.io
'''
from itertools import chain

from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class WhereClauseRenderer:

    def __init__(self, is_primary_key_selector=False):
        # if this is a primary key selector, we must not recurse into foreign keys
        self._is_primary_key_selector = is_primary_key_selector

    # This compiler create the usual where clause, but adds statements to restrict the 'using' tables
    def render(self, mapping: Entity):

        # find clauses for unique columns
        clauses = list(chain(*[self._attribute(a, mapping.name) for a in mapping.attributes if a]))

        # must not be empty, except for primary key selector, because that may be restricted by join instead
        if not clauses and not self._is_primary_key_selector:
            raise Exception('Header must have at least one unique column')

        return ' where ' + ' and '.join(clauses)

    def _attribute(self, attribute: AbstractAttribute, alias):
        # return where clause on provided parameter for unique attribute
        if isinstance(attribute, Attribute) and attribute.unique:
            parameter_name = attribute.parameter or attribute.name

            # if there's a key, then address field in json object
            if attribute.key:
                return [f"{alias}.{attribute.name}->>'{attribute.key}' = :{parameter_name}"]
            return [f'{alias}.{attribute.name} = :{parameter_name}']

        # return where clause on provided filter, even if the field is not unique
        if isinstance(attribute, Attribute) and attribute.filter:
            # replace '$' with the attribute name, including alias
            name = f'{alias}.{attribute.name}'
            return [attribute.filter.replace('$', name)]

        if self._is_primary_key_selector:
            # a primary key selector uses join...on instead of where...and, so don't create a where clause
            return []

        # return where clause on foreign key
        if isinstance(attribute, Reference) and attribute.unique:
            # update and delete statements must restrict by foreign keys, no need to recurse
            target_alias = attribute.alias or attribute.table
            return [f'{alias}.{attribute.name} = {target_alias}.{attribute.target_name}']

        # no where clause
        return []


"""
This class provides the main database service to list tables, create mappings, retrieve table contents and to post updates.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
import logging
import re
from io import StringIO
from typing import Optional

import pandas as pd
import psycopg2
from pandas import DataFrame

from .abstract_orm import AbstractORM
from .context import cnx_context, get_metadata
from .csv_reader import CsvReader
from .db_reader import DbReader
from .diff_to_executor import DiffToExecutor
from .executor_service import ExecutorService
from .odoo.postgres_model_service import PostgresModelService
from .query_executor import OperationType
from .reporter import Reporter
from ..stml.header_renderer import HeaderRenderer
from ..stml.json_renderer import JsonRenderer
from ..stml.model import Entity
from ..stml.model_enricher import ModelEnricher
from ..stml.parameter_expander import ParameterExpander
from ..stml.sql.select_renderer import SelectRenderer
from ..stml.sql.types_renderer import TypesRenderer
from ..stml.stml_creator import StmlCreator
from ..stml.stml_merger import StmlMerger
from ..stml.stml_parser import StmlParser

_logger = logging.getLogger(__name__)


class DB:
    def __init__(self, orm_function=None):
        self._diff_to_sql = DiffToExecutor()
        # delay orm creation until needed
        self._orm_function = orm_function

    def get_tables(self, filter=None):

        cr = cnx_context.cr

        # query to get all tables, excluding views
        query = "SELECT table_name FROM information_schema.tables WHERE table_schema = 'public' AND table_type = 'BASE TABLE'"

        cr.execute(query)

        # fetch all rows
        all_tables = cr.fetchall()

        if filter is None:
            # return all tables
            unsorted = [t for t in all_tables]
        else:
            # use filter as a regular expression
            regex = re.compile(filter)
            unsorted = [t for t in all_tables if regex.search(t[0])]

        # get row counts
        with_counts = self.get_counts(unsorted)

        # return sorted(with_count)
        sorted_tables = sorted(with_counts, key=lambda e: e[0])

        # convert to dictionary for JSON serialization
        return [{'name': table_name, 'count': row_count} for table_name, row_count in sorted_tables]

    def get_counts(self, tables):
        cr = cnx_context.cr
        for table in tables:
            query = f"SELECT COUNT(*) FROM {table[0]}"
            cr.execute(query)
            row_count = cr.fetchone()[0]
            yield table[0], row_count

    def get_count(self, table_name, header, where_clause):
        # create metadata object from context
        metadata = get_metadata(cnx_context.cnx)

        if not header:
            # if header parameter is empty, then use the default header
            default_mapping = StmlCreator(PostgresModelService()).create(table_name)

            # remove columns that are not enabled
            attributes = [a for a in default_mapping.attributes if a.enabled]
            mapping = Entity(table_name, attributes)

        else:
            # parse header to build syntax tree
            mapping = ModelEnricher(metadata).enrich(StmlParser().parse_csv(table_name, header))

        query = SelectRenderer().compile_count_query(mapping, where_clause)

        cr = cnx_context.cr

        # execute query
        cr.execute(query)

        # fetch result
        result = cr.fetchone()

        return result[0]

    def get_table(self, table_name, header, where_clause=None, protocol='sql'):
        # create metadata object from context
        metadata = get_metadata(cnx_context.cnx)

        if not header:
            # if header parameter is empty, then use the default header
            default_mapping = StmlCreator(PostgresModelService()).create(table_name)

            # remove columns that are not enabled
            attributes = [a for a in default_mapping.attributes if a.enabled]
            mapping = Entity(table_name, attributes)

        else:
            # parse header to build syntax tree
            mapping = ModelEnricher(PostgresModelService()).enrich(StmlParser().parse_csv(table_name, header))

        df = DbReader(protocol).read_from_db(mapping, where_clause)

        return df, mapping

    def get_table_as_csv(self, table_name, header, where_clause, escapechar=None):
        # get table as dataframe
        df, mapping = self.get_table(table_name, header, where_clause)

        return self.convert_to_csv(df, mapping, escapechar)

    def convert_to_csv(self, df, mapping, escapechar=None):
        # need converters from types compiler
        column_names = HeaderRenderer().render_list(mapping)
        column_types = TypesRenderer().render(mapping, column_names)

        # get converters from column_types
        converters = column_types['write_csv_converters']

        # apply converters to dataframe
        for column in column_names:
            if column in converters:
                df[column] = df[column].apply(converters[column])

        # convert dataframe to csv
        return df.to_csv(index=False, escapechar=escapechar)

    def post_table_get_diff(self, table_name, header, where_clause, body, skiprows=0, nrows=None, insert=False, update=False, delete=False, execute=False, commit=False, post_script=None, context=None, orm=None):
        # create diffs and sql
        diffs, sql = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, nrows, insert, update, delete, post_script, context, orm)
        # if execute
        if execute:
            # execute sql statements
            results_tuple = self._execute_sql(sql, True, commit)
            # concatenate results
            results = [result for results in results_tuple for result in results]
            # zip row counts with diffs statements
            for diff, result in zip(diffs, results):
                diff.append(result.rowcount)

        return diffs

    def post_table_get_sql(self, table_name, header, where_clause, body, skiprows=0, nrows=None, insert=False, update=False, delete=False, execute=False, commit=False, post_script=None, context=None):
        # create diffs and sql
        _, query_executors = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, nrows, insert, update, delete, post_script, context)

        # execute sql statements, or fake if execute is false
        sqls = ExecutorService().execute_sql(query_executors, execute, commit)

        # convert sql to dataframe
        return self._convert_to_df(sqls, execute)

    def post_table_get_full_report(self, table_name, header, where_clause, body, skiprows=0, nrows=None, insert=False, update=False, delete=False, execute=False, commit=False,
                                   post_script=None, context=None, substitutions=None):

        # create orm service if function is provided
        orm: Optional[AbstractORM] = self._orm_function() if self._orm_function else None

        # create diffs and sql
        diff, query_executors = self._get_diffs_and_sql(table_name, header, where_clause, body, skiprows, nrows, insert, update, delete, post_script, context, orm=orm, substitutions=substitutions)

        # execute sql statements
        execution_results = ExecutorService().execute_sql(query_executors, execute, commit)

        # create full report
        return Reporter().create_post_report([table_name], [body], [context], execution_results, execute, commit, skiprows, nrows)

    def post_multiple_tables_get_full_report(self, table_names, header, where_clause, contents, skiprows=0, nrows=None, insert=False, update=False, delete=False, execute=False, commit=False,
                                             post_script=None, context=None, substitutions=None):
        assert len(table_names) == len(contents), f"Provide exactly one file for each table name, so {len(table_names)}, not {len(contents)}"
        assert header is None, "Header must be None when posting multiple tables"
        assert skiprows >= 1, "Skiprows must be at least 1 when posting multiple tables"
        assert post_script is None, "Post script must be None when posting multiple tables"
        assert context is not None and len(context) == len(table_names), "Provide exactly one context for each table name, not %s" % len(context or [])

        query_executors = []

        # create orm service if function is provided
        orm: Optional[AbstractORM] = self._orm_function() if self._orm_function else None

        # Iterate over tables here.
        for table_name, file_context, content in zip(table_names, context, contents):
            # decode binary content
            text_content = content.decode('utf-8')
            text_substitutions = substitutions.decode('utf-8') if substitutions else None

            # get header from first line
            header = text_content.split('\n', 1)[0]

            # create diffs and sql
            _, qe = self._get_diffs_and_sql(table_name, header, where_clause, text_content, skiprows, nrows, insert, update, delete, post_script, file_context, orm=orm, substitutions=text_substitutions)
            query_executors.extend(qe)

        # execute sql statements
        execution_results = ExecutorService().execute_sql(query_executors, execute, commit)

        # create full report
        return Reporter().create_post_report(table_names, contents, context, execution_results, execute, commit, skiprows, nrows)

    def _convert_to_df(self, sqls, showResult):
        # create empty pandas dataframe. First column contains sql
        if showResult:
            result = pd.DataFrame(columns=['rows', 'errors', 'sql'])
        else:
            result = pd.DataFrame(columns=['sql'])

        # iterate sql statements
        index = 0
        for er in sqls:
            # create dictionary with query and value_dict
            value_dict = {**er.params, 'sql': er.query}
            if showResult:
                value_dict = {**value_dict, 'rows': er.rowcount, 'errors': er.error}
            # append a new row to the bottom result with values from value dictionary using concat
            result = pd.concat([result, pd.DataFrame(value_dict, index=[index])], ignore_index=True)
            index += 1

            # append dependent query result
            dep_er = er.dependent_execution_result
            if dep_er:
                value_dict = {**dep_er.params, 'sql': dep_er.query}
                if showResult:
                    value_dict = {**value_dict, 'rows': dep_er.rowcount}
                result = pd.concat([result, pd.DataFrame(value_dict, index=[index])], ignore_index=True)
                index += 1

        # force pandas to not convert int columns to float if they contain NaNs
        result = result.convert_dtypes()

        # return dataframe
        return result

    def post_table_get_summary(self, table_name, header, where_clause, body, skiprows=0, nrows=None, insert=False, update=False, delete=False, execute=False, commit=False):
        pass

    def _get_diffs_and_sql(self, table_name, header, where_clause, body, skiprows, nrows, insert, update, delete, post_script, context, orm: Optional[AbstractORM] = None, substitutions: str=None):

        # if header is empty and skiprows is larger than 0, then take the first line as header
        if not header and skiprows > 0:
            # get header from first line
            header = body.split('\n', 1)[0]

        # assert that we have a header
        assert header, "Header is required, either as a parameter or as the first line in the body with skiprows set to 1"

        # parse header to build syntax tree
        mapping = ModelEnricher(PostgresModelService()).enrich(StmlParser().parse_csv(table_name, header))

        # prepare substitutions map
        substitutions_map = self._create_substitutions_map(substitutions)

        # expand parameter placeholders
        mappings = ParameterExpander().expand(mapping, substitutions_map)

        diffs = None
        sqls = []

        # iterate over mappings
        for mapping in mappings:

            # read dataframe from request first, so we can give feedback on errors in the request
            df_request = CsvReader().read_from_request(mapping, body, skiprows, nrows, post_script, substitutions_map)

            # read dataframe from DB
            df_db = DbReader().read_from_db(mapping, where_clause, set_index=True)

            # todo: remove the need to return diffs
            diffs = self._compare(df_request, df_db, insert, update, delete)

            # create sql statements and parameters
            sqls.extend(self._diff_to_sql.diff_executor(mapping, diffs, context, orm))

        return diffs, sqls

    def _compare(self, df_request, df_db, insert, update, delete):
        # remove columns with empty names. Don't do this when reading from DB, because in get_table request we also want empty columns
        df_db = df_db.drop(columns=[''], errors='ignore')

        # find rows to insert
        inserted_indices = set(df_request.index.values).difference(df_db.index.values)
        inserts = df_request[df_request.index.isin(inserted_indices)]
        inserts.reset_index(inplace=True)

        # move __line__ to become the left most column after index reset
        inserts = self._move_line_to_front(inserts)

        # find rows to delete
        deleted_indices = set(df_db.index.values).difference(df_request.index.values)
        deletes = df_db[df_db.index.isin(deleted_indices)]
        deletes.reset_index(inplace=True)

        # find rows to update. Left is from request, right is from database
        left = df_request[~df_request.index.isin(inserted_indices)]
        right = df_db[~df_db.index.isin(deleted_indices)]

        # Sort the right DataFrame to match the index of the left DataFrame
        right_df_sorted = right.reindex(left.index)

        # get columns that are in left but not in right, such as __line__ and API filled columns
        drop_column_names = set(left.columns).difference(right_df_sorted.columns)

        # drop these columns from left
        left_no_line = left.drop(columns=drop_column_names, errors='ignore')

        # get mask of cells that are N/A or '' in left and right
        mask = (left_no_line.isna() | left_no_line.eq('')) & (right_df_sorted.isna() | right_df_sorted.eq(''))

        # replace N/A or '' with None using mask to ignore differences in N/A or ''
        left_no_line = left_no_line.mask(mask, None)
        right_df_sorted = right_df_sorted.mask(mask, None)

        # compare remaining rows
        updates = left_no_line.compare(right_df_sorted)

        # restore dropped columns from left into updates
        for column_name in drop_column_names:
            updates[column_name] = left.loc[updates.index][column_name]

        # reset index for updates to match the insert and deletes
        updates.reset_index(inplace=True)

        # move __line__ to become the left most column
        updates = self._move_line_to_front(updates)

        # return result
        return inserts if insert else DataFrame(), updates if update else DataFrame(), deletes if delete else DataFrame()

    def _move_line_to_front(self, df):
        # move __line__ to become the left most column. This has no real purpose, but it makes the dataframes more readable
        # this must also work with the multi-index dataframes coming from the compare function
        # pop __line__ column, don't use drop because that triggers a warning
        line = df.pop('__line__')
        # insert __line__ column as first column
        df.insert(0, '__line__', line)
        return df

    def get_header_json(self, table_name, header=None):
        # get mapping, merged if needed
        mapping = self._get_header_mapping(table_name, header)

        # compile into header
        json_header = JsonRenderer().render_json(mapping)

        return json_header

    def get_header_csv(self, table_name, header=None):
        # get mapping, merged if needed
        mapping = self._get_header_mapping(table_name, header)

        # compile into header
        csv_header = HeaderRenderer().render_csv(mapping)

        return csv_header

    def _get_header_mapping(self, table_name, header=None):
        # get a MetaData instance from context
        metadata = get_metadata(cnx_context.cnx)

        # create default mapping from table
        mapping = StmlCreator(PostgresModelService()).create(table_name)

        if not header:
            # return full mapping
            return mapping

        # parse header to build syntax tree
        parsed_header = ModelEnricher(PostgresModelService()).enrich(StmlParser().parse_csv(table_name, header))

        # merge headers
        merged_mapping = StmlMerger().merge(mapping, parsed_header)

        return merged_mapping



    def set_context(self, url, password):
        # create psycopg2 connection
        cnx = psycopg2.connect(url, password=password)
        # set connection and cursor in context
        cnx_context.cnx = cnx
        cnx_context.cr = cnx.cursor()


    def _get_selected_columns(self, table_name, all_columns, header):
        # get cnx from context
        cnx = cnx_context.cnx

        # parse header to build syntax tree
        mapping = ModelEnricher(PostgresModelService()).enrich(StmlParser().parse_csv(table_name, header))

        # reconstruct original header from tree
        column_keys = HeaderRenderer().render_list(mapping)

        # create result list for selected and unselected columns
        result_columns = []

        # initiate list for unselected columns with all default columns
        unselected_columns = all_columns.copy()

        # iterate header columns to retain order of selected items
        for key in column_keys:

            # find column with matching key
            column = next((c for c in unselected_columns if c['key'] == key), None)

            # raise error if if column is not found,
            if column is None:
                raise ValueError(f'Column {key} not found in table {table_name}')

            # set selected is true
            column['selected'] = True

            # remove from unselected columns and add to selected columns
            unselected_columns.remove(column)
            result_columns.append(column)

        # add remaining unselected columns to selected columns
        result_columns.extend(unselected_columns)

        return result_columns

    def _create_substitutions_map(self, substitutions):
        # convert substitutions table into a dictionary
        if not substitutions:
            # return None to indicate no substitutions were provided
            return None

        # read substitutions from binary string into a DataFrame
        df_substitutions = pd.read_csv(StringIO(substitutions), dtype=str, na_filter=False)

        # convert to dictionary, first column is the domain
        map = {row.strip(): {} for row in df_substitutions.iloc[:, 0].unique()}

        # iterate rows, second column is the name, third column is the substitution for the name
        for row in df_substitutions.itertuples(index=False):
            map[row[0].strip()][row[1].strip()] = row[2].strip()

        return map



"""
This class generates delete statements based on a mapping.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from typing import List

from stimula.stml.model import Entity, AbstractAttribute, Reference
from stimula.stml.sql.foreign_where_renderer import ForeignWhereClauseRenderer
from stimula.stml.sql.insert_renderer import ReturningClauseRenderer
from stimula.stml.sql.where_renderer import WhereClauseRenderer


class DeleteRenderer:
    """
        delete from c
        using c
        left join b on c.c0 = b.b0
        where c.c1 = :c1 and b.b1 = :b1;
    """

    def render(self, mapping: Entity):
        using_clause = UsingClauseRenderer().render(mapping)
        where_clause = WhereClauseRenderer().render(mapping)
        foreign_where_clause = ForeignWhereClauseRenderer(False, True).render(mapping)
        # in a delete statement, the foreign where clause always needs an 'and'
        foreign_where_clause = ' and ' + foreign_where_clause if foreign_where_clause else ''
        # returning clause is needed if we need to delete from an extension table
        returning_clause = ReturningClauseRenderer().render(mapping)

        result = f'delete from {mapping.name}{using_clause}{where_clause}{foreign_where_clause}{returning_clause}'

        return result


class UsingClauseRenderer:
    # this class is similar to FromClauseCompiler, but it only uses unique columns
    def render(self, mapping: Entity):
        # get clauses for unique foreign keys
        clauses = [self._attribute(a) for a in mapping.attributes if a.unique and isinstance(a, Reference)]

        if not clauses:
            return ''
        return f' using ' + ', '.join(clauses)

    def _attribute(self, attribute: Reference) -> str:

        # add target table
        from_clause = attribute.table

        # add alias
        target_alias = attribute.alias or attribute.table
        if target_alias != attribute.table:
            from_clause += f' as {target_alias}'

        # recurse
        return from_clause + ''.join(self._nested_attributes(attribute.attributes, target_alias))

    def _nested_attributes(self, attributes: List[AbstractAttribute], source_alias) -> List[str]:
        # get attributes that contains a foreign key
        return [self._nested_attribute(a, source_alias) for a in attributes if isinstance(a, Reference)]

    def _nested_attribute(self, attribute: Reference, source_alias) -> str:

        # if not root, then assume we need a left join. This may require a modifier at some point.
        from_clause = ' left join '

        # add target table
        from_clause += attribute.table

        # add alias
        target_alias = attribute.alias or attribute.table
        if target_alias != attribute.table:
            from_clause += f' as {target_alias}'

        from_clause += f' on {source_alias}.{attribute.name} = {target_alias}.{attribute.target_name}'

        # recurse
        return from_clause + ''.join(self._nested_attributes(attribute.attributes, target_alias))



"""
This class renders a mapping into a csv-style file header format

Author: Romke Jonker
Email: romke@stml.io
"""
from typing import List

from stimula.stml.model import AbstractAttribute, Attribute, Reference


class JsonRenderer:
    # attributes to include as UI directives
    UI_ATTRIBUTES = ['enabled', 'primary_key', 'unique', 'in_use', 'default']
    # attributes to exclude from the modifiers list
    NON_MODIFIER_ATTRIBUTES = ['name', 'type', 'table', 'target_name', 'attributes', 'enabled', 'primary_key', 'foreign_key', 'in_use', 'default']

    def render_json(self, mapping):
        result = {
            'table-name': mapping.name
        }
        # compile columns into json
        if mapping.attributes:
            result['columns'] = [self._column_json(a) for a in mapping.attributes]
        return result

    def _column_json(self, attribute: AbstractAttribute):
        key_without_modifiers, type = self._attributes([attribute])
        modifiers = self._modifiers_json(attribute)

        # add modifiers to key
        key = key_without_modifiers + self._modifiers(attribute)

        if not type:
            return {'key': key, **modifiers}
        return {'key': key, 'type': type, **modifiers}

    def _modifiers_json(self, attribute: AbstractAttribute):
        # convert to dict
        attribute_dict = attribute.to_dict()

        # copy column attributes that are UI directives
        keys = [k for k in sorted(attribute_dict.keys()) if k in self.UI_ATTRIBUTES]
        modifiers = {key.replace('_', '-'): attribute_dict[key] for key in keys}

        # foreign-key is a special case, check if the attribute is a foreign keys
        if isinstance(attribute, Reference):
            modifiers['foreign-key'] = True

        return modifiers

    def _attributes(self, attributes: List[AbstractAttribute]):
        keys, types = zip(*[self._attribute(a) for a in attributes])
        type = types[0] if len(set(types)) == 1 else None
        return ':'.join(keys), type

    def _attribute(self, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            return attribute.name, (attribute.type)

        if isinstance(attribute, Reference):
            key, type = self._foreign_key(attribute)
            return attribute.name + key, type

    def _foreign_key(self, foreign_key: Reference):
        key, type = self._attributes(foreign_key.attributes)
        return f"({key})", type

    def _modifiers(self, attribute: AbstractAttribute):
        # convert to dict
        attribute_dict = attribute.to_dict()

        # use an exclusion list, so we can add more attributes later
        modifiers = [self._modifier(key, attribute_dict[key]) for key in sorted(attribute_dict.keys()) if key not in self.NON_MODIFIER_ATTRIBUTES]
        if not modifiers:
            return ''
        return '[' + ': '.join(modifiers) + ']'

    def _modifier(self, key, value):
        # replace '_' with '-' in key
        key = key.replace('_', '-')

        # if type is not yet string, then convert to string and lowercase, so we get 'true' and not 'True'
        if not isinstance(value, str):
            value = str(value).lower()

        # need to quote if value has any of the following: ' "\'='
        if any(c in value for c in [' ', '"', "'", '=']):
            return f'{key}="{value}"'
        return f'{key}={value}'



"""
This class parses a line of text representing values and returns a list with those values.

A single position can contain a simple value, a combined value, a JSON object or a JSON array.

Author: Romke Jonker
Email: romke@rnadesign.net

Examples of permitted values:
    1. Simple values: 123, abcd
    2. Combined values: 123:abcd
    3. Double quotes: "abcd", "123", "123:abcd"
    4. Single quotes: 'abcd', '123', '123:abcd'
    5. Combined values with double quotes: "123":"abcd"
    6. Combined values with single quotes: '123':'abcd'
    7. JSON object: { "a": 1, "b": 2 }
    8. JSON array: [ 1, 2, 3, 4 ]
    9. JSON object with double quotes: "{ \"a\": 1, \"b\": 2 }"
    10. JSON array with double quotes: "[ 1, 2, 3, 4 ]"
    11. JSON object with single quotes: '{ "a": 1, "b": 2 }'
    12. JSON array with single quotes: '[ 1, 2, 3, 4 ]'
    13. Combined values with JSON: 123:{ "a": 1, "b": 2 }

"""
from .sly import Lexer, Parser
import pprint


class ValuesLexer(Lexer):
    tokens = {"FLOAT", "INTEGER", "STRING", "UNQUOTED_STRING"}

    literals = {'{', '}', '[', ']', ',', ':'}
    ignore = " \t\n"

    @_(r"[\"'].*?[\"']")
    # also accept strings in single quotes, because that's how json typically comes in a CSV file
    def STRING(self, t):
        t.value = t.value[1: -1]
        return t

    @_(r"\d+\.\d*")
    def FLOAT(self, t):
        t.value = float(t.value)
        return t

    @_(r"\d+")
    def INTEGER(self, t):
        t.value = int(t.value)
        return t

    @_(r'[^[\]{}:,"\']+')
    def UNQUOTED_STRING(self, t):
        return t


class ValuesParser(Parser):
    tokens = ValuesLexer.tokens
    start = "attributes"

    @_('attribute_value')
    def attributes(self, p):
        return [p.attribute_value]

    @_('attribute_value ":" attributes')
    def attributes(self, p):
        return [p.attribute_value] + p.attributes

    @_('value', 'UNQUOTED_STRING')
    def attribute_value(self, p):
        return p[0]

    @_('"{" members "}"')
    def object(self, p):
        return {key: value for key, value in p.members}

    @_('pair')
    def members(self, p):
        return [p.pair]

    @_('pair "," members')
    def members(self, p):
        return [p.pair] + p.members

    @_('STRING ":" value')
    def pair(self, p):
        return p.STRING, p.value

    @_('"[" elements "]"')
    def array(self, p):
        return p.elements

    @_('value')
    def elements(self, p):
        return [p.value]

    @_('value "," elements')
    def elements(self, p):
        return [p.value] + p.elements

    @_('STRING',
       'INTEGER',
       'FLOAT',
       'object',
       'array')
    def value(self, p):
        return p[0]

    def error(self, p):
        raise ValueError("Parsing error at token %s" % str(p))



"""
This class parses an STML mapping header. It takes CSV formatted columns and outputs the corresponding mapping in as a dictionary.

Author: Romke Jonker
Email: romke@stml.io

The grammar is as follows:

attribute    : name
             | reference
             | empty

name         : ID
             | ID modifiersets

reference    : ID ( attributes )
             | ID ( attributes ) modifiersets

attributes   : attributes COLON attribute
             | attribute

modifiersets : modifiersets modifierset
             | modifierset

modifierset  : [ modifiers ]

modifiers    : modifiers COLON modifier
             | modifier

modifier     : ID = ID
             | ID = VALUE

"""
import csv
import io

from .model import Attribute, Entity, Reference
from .sly import Parser, Lexer


class StmlLexer(Lexer):
    # Set of token names.   This is always required
    tokens = {ID, VALUE, QUOTED_VALUE, COLON, EQUALS, LPAREN, RPAREN, LBRACK, RBRACK}

    # String containing ignored characters between tokens
    ignore = ' \t'

    # Regular expression rules for tokens
    ID = r'[a-zA-Z_][a-zA-Z0-9_-]*'
    VALUE = r'[a-zA-Z0-9_${}]+'
    QUOTED_VALUE = r'"[^\"]+"'
    COLON = r'\:'
    EQUALS = r'\='
    LPAREN = r'\('
    RPAREN = r'\)'
    LBRACK = r'\['
    RBRACK = r'\]'


class StmlParser(Parser):
    # CSV headers that are treated as boolean
    BOOLEAN_HEADERS = ['unique', 'skip']
    _lexer = StmlLexer()

    # Get the token list from the lexer (required)
    tokens = StmlLexer.tokens

    # main parse function that starts with decoding CSV style headers
    def parse_csv(self, table_name, header):

        # an empty string can either be an empty header or a header with a single empty column. Let's return an empty header in that case.
        if header.strip() == '':
            return Entity(table_name)

        # decode CSV style header
        csv_file = io.StringIO(header)

        # Create a CSV reader object
        csv_reader = csv.reader(csv_file, delimiter=',', quotechar='"', skipinitialspace=True)

        # Parse the single line into a list of columns, we can't unescape in a single parsing step
        cells = next(csv_reader)

        # parse the cells, keep a cell counter to raise a more informative error message
        attributes = []
        for i, cell in enumerate(cells):
            try:
                attributes.append(self.parse(StmlParser._lexer.tokenize(cell)))
            except Exception as e:
                raise ValueError(f"Error parsing cell {i + 1} '{cell}': {str(e)}")

        # enable top level attributes STML editor
        for a in attributes:
            if a:
                a.enabled = True

        return Entity(table_name, attributes)

    @_('name')
    def attribute(self, p):
        return Attribute(p.name[0], **self._fix_modifier_names(p.name[1]))

    @_('reference')
    def attribute(self, p):
        return Reference(p.reference[0], p.reference[1], **self._fix_modifier_names(p.reference[2]))

    @_('')
    def attribute(self, p):
        return None

    @_('ID')
    def name(self, p):
        return (p.ID, {})

    @_('ID modifiersets')
    def name(self, p):
        return (p.ID, p.modifiersets)

    @_('ID LPAREN attributes RPAREN')
    def reference(self, p):
        return (p.ID, p.attributes, {})

    @_('ID LPAREN attributes RPAREN modifiersets')
    def reference(self, p):
        return (p.ID, p.attributes, p.modifiersets)

    @_('attributes COLON attribute')
    def attributes(self, p):
        return p.attributes + [p.attribute]

    @_('attribute')
    def attributes(self, p):
        return [p.attribute]

    @_('modifiersets modifierset')
    def modifiersets(self, p):
        return p.modifiersets | p.modifierset

    @_('modifierset')
    def modifiersets(self, p):
        return p.modifierset

    @_('LBRACK modifiers RBRACK')
    def modifierset(self, p):
        return p.modifiers

    @_('modifiers COLON modifier')
    def modifiers(self, p):
        return p.modifiers | p.modifier

    @_('modifier')
    def modifiers(self, p):
        return p.modifier

    @_('ID EQUALS ID')
    @_('ID EQUALS VALUE')
    def modifier(self, p):
        key = p[0]
        if key in self.BOOLEAN_HEADERS:
            # boolean modifier
            value = str(p[2]).lower() == 'true'
        else:
            value = p[2]

        return {key: value}

    @_('ID EQUALS QUOTED_VALUE')
    def modifier(self, p):
        key = p[0]
        # strip quotes
        value = p[2][1:-1]

        return {key: value}

    def _fix_modifier_names(self, modifiers):
        return {k.replace('-', '_'): v for k, v in modifiers.items()}

    def error(self, token):
        msg = 'Parse error'
        if token:
            # if token has value attribute
            value = getattr(token, 'value', None)
            if value:
                msg += f': encountered \'{value}\''

            # if token has index attribute
            index = getattr(token, 'index', 0)
            if index:
                msg += f' at index {index}'

        # fail fast
        raise ValueError(msg)



"""
This class assists users in creating a default mapping based on an existing data model.
It assumes an Odoo-flavored data model by excluding columns that are typically used by Odoo for internal
purposes.

Author: Romke Jonker
Email: romke@stml.io
"""
from typing import List

from sqlalchemy import UniqueConstraint, PrimaryKeyConstraint
from sqlalchemy.dialects.postgresql import JSONB

from stimula.service.model_service import ModelService
from stimula.stml.model import Entity, Attribute, Reference, AbstractAttribute

ODOO_SYSTEM_COLUMNS = ['create_date', 'create_uid', 'write_date', 'write_uid']


class StmlCreator:

    # take metadata and cursor as instance variables
    def __init__(self, model_service: ModelService):
        self._model_service: ModelService = model_service

    def create(self, table_name):
        # Get the table object
        table = self._model_service.get_table(table_name)

        if table is None:
            raise ValueError('Table not found: ' + table_name)

        return Entity(table.name, self._attributes(table))

    def _attributes(self, table):
        # get in use columns once, in a single query
        in_use_columns = self._model_service.get_non_empty_columns(table)

        # get all columns
        attributes = [self._attribute(c, in_use_columns) for c in table.columns]

        # We must have at least one column enabled. If no columns are enabled, then enable all columns.
        if not any([a.enabled for a in attributes]):
            for a in attributes:
                a.enabled = True

        # Make sure we have at least one unique column
        self._enforce_unique_column(attributes)

        # sort columns
        sorted_columns = self._sort_columns(attributes)

        return sorted_columns

    def _enforce_unique_column(self, attributes: List[AbstractAttribute]):
        # We must have at least one unique key. If no unique keys are enabled, then find the primary key and mark is as unique, default and enabled.
        if any([a.unique for a in attributes]):
            # fine, we have a unique column
            return
        for a in attributes:
            if a.primary_key:
                a.unique = True
                a.default = True
                a.enabled = True

    def _attribute(self, column, in_use_columns):
        if not column.foreign_keys:
            # create attribute
            attribute = Attribute(column.name, type=str(column.type).lower())
        else:
            # create foreign key attribute
            attribute = self._foreign_key(column)

        # set in-use attribute
        if column.name in in_use_columns:
            attribute.in_use = True

        # get constraints
        constraints = self._get_constraints(column)

        # set primary key and unique attributes
        for constraint in constraints:
            if constraint.__class__ == PrimaryKeyConstraint:
                attribute.primary_key = True
            elif self._is_applicable_unique_constraint(constraint):
                attribute.unique = True

        # set default attribute
        default = not attribute.primary_key and not attribute.name in ODOO_SYSTEM_COLUMNS and attribute.in_use
        if default:
            attribute.default = True
            attribute.enabled = True

        return attribute

    def _get_constraints(self, column):
        constraints = [constraint for constraint in column.table.constraints if column in list(constraint.columns)]
        return constraints

    def _foreign_key(self, column) -> Reference:
        foreign_keys = column.foreign_keys
        if len(foreign_keys) != 1:
            raise ValueError(f"Expected 1 foreign key, found {len(foreign_keys)}")
        foreign_key = list(foreign_keys)[0]
        table = foreign_key.column.table
        target_name = foreign_key.column.name
        return Reference(column.name, table=table.name, target_name=target_name, attributes=self._unique_columns(table))

    def _unique_columns(self, table) -> List[AbstractAttribute]:
        # find applicable constraints
        unique_constraints = [c for c in table.constraints if self._is_applicable_unique_constraint(c)]
        primary_key_constraints = [c for c in table.constraints if c.__class__ == PrimaryKeyConstraint]

        if len(unique_constraints) > 0:
            # use unique columns from first unique constraint
            columns = unique_constraints[0].columns
        elif len(primary_key_constraints) > 0:
            # fall back on primary key columns
            columns = primary_key_constraints[0].columns
        else:
            columns = []

        return [self._unique_column(c) for c in columns]

    def _is_applicable_unique_constraint(self, c):
        # must be a unique constraint
        if c.__class__ != UniqueConstraint:
            return False

        # don't set jsonb columns as unique, because the resulting dict is not hashable
        if any([isinstance(column.type, JSONB) for column in c.columns]):
            return False

        # else it's applicable
        return True

    def _unique_column(self, column) -> AbstractAttribute:
        if not column.foreign_keys:
            return Attribute(column.name, type=str(column.type).lower())
        else:
            return self._foreign_key(column)

    def _sort_columns(self, attributes: List[AbstractAttribute]) -> List[AbstractAttribute]:
        # get primary key columns
        primary_key_columns = [a for a in attributes if a.primary_key]

        # get unique columns, exclude primary keys
        unique_columns = [a for a in attributes if not a.primary_key and a.unique]

        # get other columns
        other_columns = [a for a in attributes if not a.primary_key and not a.unique]

        return sorted(primary_key_columns, key=lambda a: a.name) \
            + sorted(unique_columns, key=lambda a: a.name) \
            + sorted(other_columns, key=lambda a: a.name)



"""
This script provides a command-line interface (CLI) for interacting with the Stimula API.
It supports both local and remote invocations, allowing users to execute a variety of commands such as
authentication, listing tables, retrieving mappings, counting records, fetching tables, and posting data.

Author: Romke Jonker
Email: romke@rnadesign.net

Usage:
python cli.py <command> [options]

Commands:
- auth: Authenticate, retrieve a token and store it in a local file.
- list: List tables and their record counts.
- mapping: Retrieve mapping for a specified filter.
- count: Count records in a table based on specified criteria.
- get: Fetch a table with specified mapping and query parameters.
- post: Post data to a table from a specified file.

Options:
-h, --help        Print help message
-r, --remote      Remote API URL
-H, --host        Database host (default: localhost)
-P, --port        Database port (default: 5432)
-d, --database    Database name
-u, --user        Database username
-p, --password    Password
-k, --key         Secret key
-T, --token       Authentication token
-t, --table       Table name or filter
-q, --query       Query clause
-m, --mapping     Mapping header
-f, --file        Path to the file to post (required for post command)
-s, --skip        Number of rows to skip (default: 1)
-n, --nrows       Number of rows to read (default: all)
-e, --enable      Enable flags (I, U, D, E, C)
-F, --format      Response format (choices: diff, sql)
-v, --version     Print version information
-V, --verbose     Increase output verbosity
-M, --transpose   Transpose the mapping
-x, --execute     Script to execute on post
"""

import argparse
import getpass
import json
import os
import sys
from importlib.metadata import version
from io import StringIO

import pandas as pd

from stimula.cli import local, remote
from stimula.cli.anonymizer import Anonymizer
from stimula.cli.file_source import FileSource
from stimula.cli.google_source import GoogleSource, google_authenticate
from stimula.service.query_executor import OperationType


class StimulaCLI:
    def main(self):
        # parse command line arguments
        args = self.parse_args()

        try:
            # execute command
            self.execute_command(args)

            return 0
        except Exception as e:
            if args.verbose:
                # print message with stack trace
                raise e
            else:
                # print first line, to avoid the remote stacktrace
                first_line = str(e).split('\n')[0]
                print(f'Error: {first_line}')

            return 1

    def parse_args(self):
        parser = argparse.ArgumentParser(description='stimula - The STML CLI')
        parser.add_argument('command', help='Command to execute', choices=['auth', 'list', 'mapping', 'count', 'get', 'post', 'transpose', 'google', 'anonymize'])
        parser.add_argument('-r', '--remote', help='Remote API URL')
        parser.add_argument('-H', '--host', help='Database host', default='localhost')
        parser.add_argument('-P', '--port', help='Database port', type=int, default=5432)
        parser.add_argument('-d', '--database', help='Database name')
        parser.add_argument('-u', '--user', help='Database username')
        parser.add_argument('-p', '--password', help='Password')
        parser.add_argument('-k', '--key', help='Secret key')
        parser.add_argument('-T', '--token', help='Authentication token')
        parser.add_argument('-t', '--tables', nargs='+', help='One or more table names or a table name filter')
        parser.add_argument('-q', '--query', help='Query clause')
        parser.add_argument('-m', '--mapping', help='Mapping header')
        parser.add_argument('-f', '--files', nargs='+', help='One or more paths to the files to post')
        parser.add_argument('-s', '--skip', help='Number of rows to skip', type=int, default=1)
        parser.add_argument('-n', '--nrows', help='Number of rows to read', type=int, default=None)
        parser.add_argument('-b', '--block_size', help='Number of rows to send at once', type=int, default=None)
        parser.add_argument('-F', '--format', help='Response format', choices=['diff', 'sql', 'full'], default='full')
        parser.add_argument('-v', '--version', action='version', version=version('stimula'))
        parser.add_argument('-V', '--verbose', action='store_true', help='Increase output verbosity')
        parser.add_argument('-M', '--transpose', action='store_true', help='Transpose the mapping')
        parser.add_argument('-x', '--execute', help='Script to execute on post')
        parser.add_argument('-c', '--context', nargs='+', help='Free text to match the query results, usually a source file name')
        parser.add_argument('-G', '--google_auth', nargs='?', help='Optional path of Google credentials file', const='client_secret.json')
        parser.add_argument('-g', '--google_sheet', nargs='?', help='ID of Google Sheets document')
        parser.add_argument('-a', '--audit', action='store_true', help='Print audit trail')
        parser.add_argument('-I', '--insert', action='store_true', help='Enable INSERT operations')
        parser.add_argument('-U', '--update', action='store_true', help='Enable UPDATE operations')
        parser.add_argument('-D', '--delete', action='store_true', help='Enable DELETE operations')
        parser.add_argument('-C', '--commit', action='store_true', help='Commit transaction')
        args = parser.parse_args()
        return args

    def execute_command(self, args):

        if args.command == 'transpose':
            # transpose stdin to stdout and exit
            self._transpose_stdin_stdout()
            return

        if args.command == 'anonymize':
            # transpose stdin to stdout and exit
            self._anonymize_stdin_stdout()
            return

        if args.command == 'google':
            assert args.google_auth, 'No Google credentials file provided. Use --google-auth or -G to provide Google credentials json file.'
            google_authenticate(args.google_auth)

        # if this is not auth, and no token provided, then try to read from local file
        if args.command != 'auth' and not args.token:
            self._read_token_from_file(args)

        if args.remote:
            # if remote is specified, use remote invoker
            invoker = remote.Invoker(args.remote)
        else:
            # otherwise, read key from environment if not provided as argument
            if not args.key:
                args.key = os.getenv('STIMULA_KEY')

            # if key is still not provided, raise an error
            assert args.key, 'Secret key must be provided for local connection. Either connect to remote API, or set --key argument, or STIMULA_KEY environment variable'

            # use local invoker
            invoker = local.Invoker(args.key, args.host, args.port)

        # if auth request or no token provided
        if args.command == 'auth' or not args.token:
            # authenticate and set token
            self._authenticate(args, invoker)

        # validate token and set connection context
        invoker.set_context(args.token)

        # check if mapping is provided as file name, then use the first line
        self._read_mapping_from_file(args)

        if args.transpose:
            # transpose mapping
            self._transpose_mapping(args)

        # use only first line of mapping
        if args.mapping:
            args.mapping = args.mapping.splitlines()[0]

        # execute command
        if args.command == 'auth':
            if args.remote:
                print(f'Connected to {args.remote} as {args.user}.')
            else:
                print(f'Connected as {args.user}.')
        elif args.command == 'list':
            filter = args.tables[0] if args.tables else None
            tables = invoker.list(filter)
            # print name and count of tables
            for table in tables:
                if args.verbose or table["count"] > 0:
                    print(f'{table["name"]}: {table["count"]}')
        elif args.command == 'mapping':
            assert args.tables and len(args.tables) == 1, 'One table name must be provided using -t or --table flag.'
            mapping = invoker.mapping(args.tables[0])
            print(mapping)
        elif args.command == 'count':
            assert args.tables and len(args.tables) == 1, 'One table name must be provided using -t or --table flag.'
            count = invoker.count(args.tables[0], args.mapping, args.query)
            print(count)
        elif args.command == 'get':
            assert args.tables and len(args.tables) == 1, 'One table name must be provided using -t or --table flag.'
            table = invoker.get_table(args.tables[0], args.mapping, args.query)
            print(table)
        elif args.command == 'post':

            # we need at least one of the flags I, U, D,  enabled
            assert args.insert or args.update or args.delete, 'At least one of the flags I, U, D must be enabled. Otherwise, there\'s nothing to do.'

            # read data from Google Sheets or from file
            if args.google_sheet:
                source = GoogleSource(args.google_sheet)
            else:
                source = FileSource()

            # don't specify nrows or block_size  together with delete
            assert not (args.nrows and args.delete), 'Cannot specify --nrows (n) and --delete (D) together. That could result in unexpected deletes.'
            assert not (args.block_size and args.delete), 'Cannot specify --block_size (b) and --delete (D) together. That could result in unexpected deletes.'

            # read files from disk, stdin or google sheets. Also evaluate table and context
            files, tables, context, substitutions = source.read_files(args.files, args.tables, args.context)

            if args.mapping is None or args.mapping == '':
                assert args.skip > 0, 'No mapping provided and skip is zero. Specify a mapping using the --mapping or -m flag, or provide a file with a header row and --skip > 0.'
                assert not args.transpose, 'Cannot transpose mapping when reading header from data file.'
                # leave it to the server to use first line of contents as mapping

            # if verbose, print mapping
            if args.verbose and args.mapping:
                print(f'Mapping: {args.mapping}')

            result = invoker.post_table(tables, args.mapping, args.query, files,
                                        skiprows=args.skip,
                                        nrows=args.nrows,
                                        block_size=args.block_size,
                                        insert=args.insert,
                                        update=args.update,
                                        delete=args.delete,
                                        execute=True,
                                        commit=args.commit,
                                        format=args.format,
                                        post_script=args.execute,
                                        context=context,
                                        substitutions=substitutions)

            print(self._create_report(result, args.audit, args.verbose))

    def _authenticate(self, args, invoker):
        # assert that database and username are provided if we don't have a token
        if not args.token:
            # database is not required for single db Odoo instance
            assert args.user, 'Username must be provided for authentication.'

        # if we have a token, then use it for default database and username
        if args.token:
            # get database and username from token
            database, user = invoker.get_database_and_username(args.token)
            # default to token values if not provided
            args.database = args.database or database
            args.user = args.user or user

        # ask for password if not provided
        if not args.password:
            # hide password input
            if args.database:
                args.password = getpass.getpass(f'Enter password for {args.user}@{args.database}: ')
            else:
                args.password = getpass.getpass(f'Enter password for {args.user}: ')

        # authenticate and set token
        args.token = invoker.auth(args.database, args.user, args.password)

        # write token to local file
        self._write_token_to_file(args.token, args.remote)

    def _transpose_stdin_stdout(self):
        # validate we have stdin
        assert not sys.stdin.isatty(), 'No input provided, use piping to provide input.'
        # read dataframe from stdin
        df = pd.read_csv(sys.stdin, header=None)
        # transpose the dataframe and write to stdout
        df.T.to_csv(sys.stdout, header=False, index=False)

    def _anonymize_stdin_stdout(self):
        # validate we have stdin
        assert not sys.stdin.isatty(), 'No input provided, use piping to provide input.'
        # anonymize the dataframe
        Anonymizer().anonymize(sys.stdin, sys.stdout)

    def _read_token_from_file(self, args):
        # check if .stimula_token file exists
        if os.path.exists('.stimula_token'):
            try:
                # read json from .stimula_token file
                with open('.stimula_token', 'r') as file:
                    data = json.load(file)
                    args.token = data.get('token')
                    args.remote = data.get('remote')
            # catch any errors,
            except Exception:
                print('Error reading token from file. Please re-authenticate.')
                pass

    def _write_token_to_file(self, token, remote):
        data = {'token': token}
        if remote:
            data['remote'] = remote
        # write token to local file
        with open('.stimula_token', 'w') as file:
            json.dump(data, file, indent=4)

    def _read_mapping_from_file(self, args):
        # check if args.mapping is an existing file name
        if args.mapping and os.path.exists(args.mapping):
            with open(args.mapping, 'r') as file:
                # read the whole file, it may be a transposed mapping
                args.mapping = file.read()

    def _transpose_mapping(self, args):
        assert args.mapping, 'Mapping must be provided when transposing.'
        # get string from args.mapping, remove lines starting with #
        mapping = '\n'.join(line for line in args.mapping.splitlines() if not line.startswith('#'))
        # use pandas to read the mapping string into a dataframe
        df = pd.read_csv(StringIO(mapping), header=None)
        # transpose the dataframe, dispose all but the first row and convert to csv
        args.mapping = df.T.head(1).to_csv(header=False, index=False)

    def _create_report(self, result, audit, verbose):
        if audit:
            return self._report_audit(result)
        elif verbose:
            return self._report_verbose(result)
        else:
            return self._report_summary(result)

    def _report_summary(self, result):
        summary = result.get('summary', {})
        total = summary.get('total', {})

        # report summary
        report = ''

        if 'rows' in summary:
            rows = summary.get("rows")
            report += f'Rows read:  {rows}\n'
            if rows == 0:
                return report

        operations = total.get("operations", 0)
        report += f'Operations: {operations}\n'
        if operations == 0:
            return report

        success = total.get('success', {})
        report += f'Success:    {success}\n'

        failed = total.get('failed', {})
        if failed > 0:
            report += f'Failed:     {failed}\n'

        error_rows = [row for row in result.get('rows', []) if not row.get('success', False)]
        error_report = '\n'.join([self._report_row(row, False) for row in error_rows])
        if error_report:
            report += f'{error_report}\n'

        if summary.get('commit', False):
            report += 'Transaction committed\n'
        else:
            report += 'Specify --commit (-C) to commit transaction.\n'

        return report

    def _report_verbose(self, result):

        summary = result.get('summary', {})
        total = summary.get('total', {})

        # report summary
        report = ''

        report += f'Rows read:  {summary.get("rows", 0)}\n'
        report += f'Operations: {total.get("operations", 0)}\n'
        success = summary.get('success', {})
        report += f'Success     {total.get("success", 0)} (insert: {success.get("insert", 0)}, update: {success.get("update", 0)}, delete: {success.get("delete", 0)})\n'
        failed = summary.get('failed', {})
        report += f'Failed      {total.get("failed", 0)} (insert: {failed.get("insert", 0)}, update: {failed.get("update", 0)}, delete: {failed.get("delete", 0)})\n'

        error_rows = [row for row in result.get('rows', []) if not row.get('success', False)]
        error_report = '\n'.join([self._report_row(row, True) for row in error_rows])

        if error_report:
            report += f'Errors:\n{error_report}\n'

        if summary.get('commit', False):
            report += 'Transaction committed\n'
        else:
            report += 'Specify --commit (-C) to commit transaction.\n'

        return report

    def _report_audit(self, result):
        # return result with indented json layout
        return json.dumps(result, default=self._custom_encoder, indent=2)

    def _custom_encoder(self, obj):
        if isinstance(obj, OperationType):
            # Convert OperationType Enum to string
            return repr(obj)
        elif isinstance(obj, pd.Timestamp):
            # Convert Timestamps to ISO format
            return obj.isoformat()
        # Fall back to default behavior for unknown types
        return json.JSONEncoder().default(obj)

    def _report_row(self, row, verbose):
        if not verbose:
            result = f'{row.get("context", "")}'
            # delete statements don't have a line number
            if row.get("line_number"):
                result += f':{row["line_number"]} - '
            result += f'"{row.get("error", "N/A")}" '
        else:
            result = f'{row.get("context", "")}'
            # delete statements don't have a line number
            if row.get("line_number"):
                result += f':{row["line_number"]} - '
            result += f'"{row.get("error", "N/A")}" - Query: "{row.get("query", "N/A")}" - Parameters: "{row.get("params", "N/A")}"'

        return result


def validate_flags(value):
    # validation function for enable flags
    valid_letters = set("iudec")
    # create a set from the input string, lowercase
    input_set = set(value.lower())

    if not input_set.issubset(valid_letters):
        raise argparse.ArgumentTypeError(f"Invalid combination: {value}. Only the letters I, U, D, E, and C are allowed.")

    return value


def main():
    return StimulaCLI().main()


if __name__ == '__main__':
    main()



"""
This class provides access to the database model in a Postgres database.

Author: Romke Jonker
Email: romke@stml.io
"""
import pandas as pd
from sqlalchemy import select, func

from stimula.service.context import cnx_context, get_metadata
from stimula.service.model_service import ModelService
from stimula.stml.alias_enricher import AliasEnricher
from stimula.stml.sql.select_renderer import SelectRenderer


class PostgresModelService(ModelService):
    def __init__(self):
        self._meta = None

    @property
    def meta(self):
        if self._meta is None:
            self._meta = get_metadata(cnx_context.cnx)
        return self._meta

    def get_table(self, table_name):
        table = self.meta.tables.get(table_name)
        if table is None:
            raise ValueError(f"Table '{table_name}' not found")
        return table

    def find_primary_keys(self, table):
        # return names of primary keys in table
        return [column.name for column in table.primary_key]

    def resolve_foreign_key_table(self, table, column_name):
        # get referred column
        if column_name not in table.columns:
            raise ValueError(f"Column '{column_name}' not found in table '{table.name}'")
        column = table.columns[column_name]
        # only know how to deal with a single foreign key per column
        if len(column.foreign_keys) != 1:
            # this is either an error condition or an extension relation, with the foreign key in the extension table. We can tell once we have parsed the modifiers.
            return None, column_name

        foreign_key = list(column.foreign_keys)[0]
        foreign_table = foreign_key.column.table
        foreign_column_name = foreign_key.column.name
        return foreign_table, foreign_column_name

    def get_non_empty_columns(self, table):
        # create list of column names
        column_names = [c.name for c in table.columns]
        # create column expressions
        expr = select(*[func.bool_or(column.isnot(None)) for column in table.columns])
        # execute query
        cr = cnx_context.cr
        cr.execute(str(expr))
        result = cr.fetchone()
        # zip and filter non-null columns
        result = [c[0] for c in zip(column_names, result) if c[1]]
        # return list
        return result

    def read_table(self, mapping: dict, where_clause=None):
        # get sqlalchemy engine from context
        engine = cnx_context.engine

        # create select query
        query = self._create_select_query(mapping, where_clause)

        # read dataframe from DB
        return pd.read_sql_query(query, engine)

    def _create_select_query(self, mapping, where_clause):
        # add aliases and parameter names
        aliased_mapping = AliasEnricher().enrich(mapping)

        # translate syntax tree to select query
        return SelectRenderer().render(aliased_mapping, where_clause)



"""
ModelService implementation for Odoo JSON-RPC
"""
import json
from itertools import chain
from typing import List

import requests

from stimula.service.model_service import ModelService
from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class JsonRpcModelService(ModelService):
    def __init__(self, client):
        self.client = client

    def get_table(self, name):
        tables = self.client.execute_kw('ir.model', 'search_read', [[['model', '=', name]]], {'fields': ['id', 'model', 'field_id']})
        if not tables:
            raise ValueError(f"Table '{name}' not found")
        if len(tables) > 1:
            raise ValueError(f"More than one table found with name '{name}'")
        table = tables[0]
        return Table(self.client, table['model'], table['field_id'])

    def find_primary_keys(self, table):
        columns = table.columns
        # in Odoo, if id exists, then it's the primary key.
        # TODO: implement for many-to-many relationships
        return [columns['id'].name] if 'id' in columns else []

    def resolve_foreign_key_table(self, table, column_name):
        # get referred column
        if column_name not in table.columns:
            raise ValueError(f"Column '{column_name}' not found in table '{table.name}'")
        column = table.columns[column_name]
        assert column.relation, f"Column '{column_name}' does not have a relation"
        foreign_table = self.get_table(column.relation)
        # for Odoo, assume the foreign column is 'id'
        foreign_column_name = 'id'
        return foreign_table, foreign_column_name

    def get_non_empty_columns(self, table):
        return []

    def read_table(self, mapping: Entity, where_clause=None):
        # get model name
        model = mapping.name

        # read attribute values from the model
        records = self._read_columns(model, mapping.attributes)

        # convert each row from list of tuples to dictionary, skip unresolved foreign keys
        return [{f[1]: self._join_values(f[2]) for f in r if len(f) == 3} for r in records]

    def _read_columns(self, model, attributes: List[AbstractAttribute]):

        # get attribute names
        fields = [a.name for a in attributes]

        # perform search_read on the model
        records = self.client.execute_kw(model, 'search_read', [[]], {'fields': fields})

        # resolve foreign keys for each attribute
        records_resolved = list(zip(*[self._resolve_foreign_key(records, a) for a in attributes]))

        # return resolved records as list
        return records_resolved

    def _resolve_foreign_key(self, records, attribute: AbstractAttribute):
        if isinstance(attribute, Attribute):
            return [(r['id'], attribute.name, [r[attribute.name]]) for r in records]

        if isinstance(attribute, Reference):
            # get foreign key attribute name
            fk = attribute.name

            # list all unique foreign key ids
            ids = list(set([r[fk][0] for r in records if r.get(fk)]))
            if not ids:
                return

            # read all foreign key values, recurse to resolve nested foreign keys. Join keys and values for nested foreign keys.
            fk_records = self._read_attributes(attribute.table, attribute.attributes, ids)

            # create a dictionary of resolved foreign key records
            fk_records_dict = {r[0]: r for r in fk_records}

            # return foreign key values
            records_resolved = [self.create_fk_record(fk, fk_records_dict, r) for r in records]

            return records_resolved

    def create_fk_record(self, fk, fk_records_dict, record):
        # get foreign key value
        fk_value = record.get(fk)

        # if foreign key is empty or false, return empty record
        if not fk_value:
            return (record['id'],)

        # return foreign key record
        return (record['id'], f'{fk}({fk_records_dict.get(record[fk][0])[1]})', fk_records_dict.get(record[fk][0])[2])

    def _read_attributes(self, model, attributes: List[AbstractAttribute], ids):
        # get attribute names
        fields = [a.name for a in attributes]

        # perform search_read on the model
        records = self.client.execute_kw(model, 'search_read', [[['id', 'in', ids]]], {'fields': fields})

        # resolve foreign keys for each attribute
        records_resolved = list(zip(*[self._resolve_foreign_key(records, a) for a in attributes]))

        # join keys, but don't join values yet
        joined_records = [(r[0][0], ':'.join([f[1] for f in r]), list(chain(*[f[2] for f in r]))) for r in records_resolved]

        return joined_records

    def _join_values(self, values):
        if len(values) == 0:
            return ''
        if len(values) == 1:
            return values[0]
        return ':'.join([self._quote(v) for v in values])

    def _quote(self, value):
        # no need to quote other than string
        if not isinstance(value, str):
            return value

        # no need to quote value without a colon
        if not ':' in value:
            return value

        # else quote
        return f'"{value}"'


class JsonRpcClient:
    def __init__(self, url, db, username, password):
        self.url = url
        self.db = db
        self.username = username
        self.password = password
        self._uid = None
        self._id = 0

    @property
    def id(self):
        self._id = self._id + 1
        return self._id

    @property
    def uid(self):
        if self._uid is None:
            self._uid = self.login()
        return self._uid

    def login(self):
        response = self.call('common', 'login', [self.db, self.username, self.password])
        assert response.get('result'), "Authentication failed; check credentials."
        return response["result"]

    def call(self, service, method, args):
        """Authenticate with Odoo and return session ID."""
        headers = {'Content-Type': 'application/json'}
        payload = {
            "id": self.id,
            "jsonrpc": "2.0",
            "method": "call",
            "params": {
                "service": service,
                "method": method,
                "args": args,
            }
        }

        response = requests.post(self.url, data=json.dumps(payload), headers=headers)
        assert response.status_code == 200, response.get('text', "Failed.")

        return response.json()

    def execute_kw(self, model, method, args, kwargs={}):
        args = [
            self.db,
            self.uid,
            self.password,
            model,  # Model for metadata about models
            method,  # Method to retrieve models
            args,
            kwargs
        ]
        response = self.call('object', 'execute_kw', args)
        assert 'result' in response, response.get('error', "Request failed.").get('data', "Request failed.")
        return response.get('result')


class Table:
    def __init__(self, client, name, field_id):
        self.client: JsonRpcClient = client
        self.name: str = name
        self._field_id: list = field_id
        self._columns: list = None

    _type_map = {'char': 'text', 'integer': 'numeric', 'many2one': 'integer'}

    @property
    def columns(self):
        if self._columns is None:
            # get field names and types
            columns = self.client.execute_kw('ir.model.fields', 'search_read', [[['id', 'in', self._field_id]]], {'fields': ['name', 'ttype', 'relation']})
            # columns = self.client.execute_kw('ir.model.fields', 'search_read', [[['id', 'in', self._field_id]]], {})
            self._columns = {column['name']: Column(column['name'], self._substitute_type(column['ttype']), column['relation']) for column in columns}
        return self._columns

    def _substitute_type(self, type):
        # replace Odoo types with SQL types
        return self._type_map.get(type, type)

    def __str__(self):
        return self.name


class Column:
    def __init__(self, name, type, relation):
        self.name: str = name
        self.type: str = type
        self.relation: str = relation

    @property
    def quoted_name(self):
        return self.name

    def __str__(self):
        return self.name



from datetime import datetime
from itertools import chain

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes

from stimula.service.context import cnx_context
from stimula.service.query_executor import OperationType


class Reporter:
    def create_post_report(self, tables, contents, contexts, execution_results, execute, commit, skiprows, nrows):

        summary = {
            'execute': execute,
            'commit': commit,
            'timestamp': datetime.now().isoformat()
        }

        # set username if available
        if cnx_context and hasattr(cnx_context, 'username'):
            summary['username'] = cnx_context.username

        try:
            # count number of rows in all contents.
            summary['rows'] = sum([self._count_rows(content, skiprows, nrows) for content in contents])
        except:
            # if counting fails, do nothing
            pass

        summary['total'] = {
            'operations': len(execution_results),
            'success': len([er for er in execution_results if er.success]),
            'failed': len([er for er in execution_results if not er.success]),
            'insert': len([er for er in execution_results if er.operation_type == OperationType.INSERT]),
            'update': len([er for er in execution_results if er.operation_type == OperationType.UPDATE]),
            'delete': len([er for er in execution_results if er.operation_type == OperationType.DELETE]),
        }

        # summarize successful operations
        summary['success'] = {'insert': len([er for er in execution_results if er.operation_type == OperationType.INSERT and er.success]),
                              'update': len([er for er in execution_results if er.operation_type == OperationType.UPDATE and er.success]),
                              'delete': len([er for er in execution_results if er.operation_type == OperationType.DELETE and er.success])}

        # summarize failed operations
        summary['failed'] = {'insert': len([er for er in execution_results if er.operation_type == OperationType.INSERT and not er.success]),
                             'update': len([er for er in execution_results if er.operation_type == OperationType.UPDATE and not er.success]),
                             'delete': len([er for er in execution_results if er.operation_type == OperationType.DELETE and not er.success])}

        # files
        files = [self._summarize_file(table, context, content) for table, context, content in zip(tables, contexts, contents)]

        # list execution results
        rows = list(chain(*[er.report(execute) for er in execution_results]))

        return {'summary': summary, 'files': files, 'rows': rows}

    def _summarize_file(self, table, context, content):
        return {'table': table, 'context': context, 'size': len(content), 'md5': self.md5_string(content)}

    def md5_string(self, input_string):
        # Create an MD5 hash object
        md5_hash = hashes.Hash(hashes.MD5(), backend=default_backend())

        # encode the input string, or leave if it's already byte like
        if isinstance(input_string, str):
            input_string = input_string.encode('utf-8')

            # Update the hash with the byte-encoded string
        md5_hash.update(input_string)

        # Finalize the hash and return the hexadecimal representation
        return md5_hash.finalize().hex()

        # Update the hash object with the byte-encoded string
        md5_hash.update(encoded_string)

        # Return the hexadecimal representation of the hash
        return md5_hash.hexdigest()

    def _count_rows(self, content, skiprows, nrows):
        # decode the input string, or leave if it's already a string
        if isinstance(content, bytes):
            content = content.decode('utf-8')
        return len(content.strip().split('\n')) - skiprows



"""
This class renders a mapping into an update query.

Author: Romke Jonker
Email: romke@rnadesign.net
"""
from .foreign_where_renderer import ForeignWhereClauseRenderer
from .insert_renderer import FromClauseRenderer
from .where_renderer import WhereClauseRenderer
from ..model import Entity, AbstractAttribute, Attribute, Reference


class UpdateRenderer:
    """
        update c
        set c2 = c0.c0, c3 = b.b0, c4 = b0.b0
        from c as c0, b, b as b0
        left join a on b0.b2 = a.a0
        where c.c1 = 'c4' and c0.c1 = 'c2' and b.b1 = 'b1' and a.a1 = 'a2';
    """

    def render(self, mapping: Entity):
        update_clause = UpdateClauseRenderer().render(mapping)
        from_clause = FromClauseRenderer(False).render(mapping)
        where_clause = WhereClauseRenderer().render(mapping)
        foreign_where_clause = ForeignWhereClauseRenderer(False, False).render(mapping)

        result = f'{update_clause}{from_clause}{where_clause}'
        if foreign_where_clause:
            result += ' and ' + foreign_where_clause
        return result


class UpdateClauseRenderer:
    def render(self, mapping: Entity):

        # can't update unique columns
        clauses = [self._attribute(a, mapping.name) for a in mapping.attributes if not a.unique]

        # comma separate cells
        return f'update {mapping.name} set ' + ', '.join(clauses)

    def _attribute(self, attribute: AbstractAttribute, table_name: str):
        if isinstance(attribute, Attribute):
            # if key is set, then update json field
            if attribute.key:
                # use table_name to disambiguate input for jsonb_set. Coalesce is used to update existing empty json fields.
                return f"{attribute.name} = jsonb_set(COALESCE({table_name}.{attribute.name}, '{{}}'::jsonb), '{{{attribute.key}}}', to_jsonb(:{attribute.parameter}::text))"

            return f'{attribute.name} = :{attribute.parameter}'

        if isinstance(attribute, Reference):
            # table names may need an alias
            target_alias = attribute.alias or attribute.table

            # if key is set, then update json field
            if attribute.key:
                # use table_name to disambiguate input for jsonb_set. Coalesce is used to update existing empty json fields.
                return f"{attribute.name} = jsonb_set(COALESCE({table_name}.{attribute.name}, '{{}}'::jsonb), '{{{attribute.key}}}', to_jsonb({target_alias}.{attribute.target_name}::text))"

            # no need to recurse
            return f'{attribute.name} = {target_alias}.{attribute.target_name}'



"""
This class compiles a mapping to enrich it with model details, such as table and attribute names.

This compiler is not thread safe, because it uses a stack to keep track of the current table and column. Make sure that
each thread has its own instance of this compiler.

Author: Romke Jonker
Email: romke@rnadesign.net

"""

from stimula.service.model_service import ModelService
from stimula.stml.model import Entity, AbstractAttribute, Attribute, Reference


class ModelEnricher:
    # CSV headers that are treated as boolean
    BOOLEAN_HEADERS = ['unique', 'skip']

    def __init__(self, model_service: ModelService):
        self._model_service: ModelService = model_service

    def enrich(self, mapping: Entity):
        # resolve table
        table = self._model_service.get_table(mapping.name)

        # find primary key, required for ORM operations
        primary_keys = self._model_service.find_primary_keys(table)

        if len(primary_keys):
            mapping.primary_key = primary_keys[0]

        # process attributes, skip None columns
        [self._attribute(table, a) for a in mapping.attributes if a]

        return mapping

    def _attribute(self, table, attribute: AbstractAttribute):

        # verify column exists
        if attribute.name not in table.columns:
            # for skip and orm-only modifiers, we don't need the column to exist
            if not attribute.skip and not attribute.orm_only:
                raise ValueError(f"Column '{attribute.name}' not found in table '{table}'")
        else:
            column = table.columns[attribute.name]
            type = str(column.type).lower()
            if type == 'jsonb' and isinstance(attribute, Attribute) and attribute.key is not None:
                # represent jsonb column with key modifier as a string
                attribute.type = 'text'
            else:
                attribute.type = type

        # resolve foreign key if needed
        if isinstance(attribute, Reference):
            self._foreign_key(table, attribute)

    def _foreign_key(self, table, reference: Reference):

        # resolve foreign key table and column
        target_table, target_column_name = self._model_service.resolve_foreign_key_table(table, reference.name)

        if target_table is None:
            # foreign key not found, this is either an error condition or an extension relation, with the foreign key in the extension table.
            target_table, target_column_name = self._extension_foreign_key(reference)

            # mark the foreign key as extension
            reference.extension = True

        reference.table = target_table.name
        reference.target_name = target_column_name

        # recurse attributes
        [self._attribute(target_table, p) for p in reference.attributes]

    def _extension_foreign_key(self, reference: Reference):
        # find foreign table name in modifiers, or default to Odoo's ir_model_data table
        table_name = reference.table or 'ir_model_data'

        # resolve the table
        table = self._model_service.get_table(table_name)

        # find referred column in cell, or default to Odoo's res_id column
        column_name = reference.target_name or 'res_id'
        if column_name not in table.columns:
            raise ValueError(f"Column '{column_name}' not found in table '{table}'")

        # Odoo's ir_model_data requires a qualifier name, verify that. For other extension tables, such as ir_attachment, the qualifier is optional.
        if table_name == 'ir_model_data' and not reference.qualifier:
            raise ValueError(f"Column '{reference.name}' is an Odoo external ID column, but no 'qualifier' is specified in modifiers")

        return table, column_name



"""
This script provides the concrete ExecutorCreator implementation to insert, update and delete using SQL.

Author: Romke Jonker
Email: romke@stml.io
"""

from .executor_creator import ExecutorCreator
from .query_executor import SimpleQueryExecutor, DependentQueryExecutor, OperationType
from ..stml.header_renderer import HeaderRenderer
from ..stml.model import Entity, Reference, Attribute
from ..stml.sql.delete_renderer import DeleteRenderer
from ..stml.sql.insert_renderer import InsertRenderer, ReturningClauseRenderer
from ..stml.sql.update_renderer import UpdateRenderer


class InsertSqlCreator(ExecutorCreator):
    def __init__(self):
        super().__init__()
        self.operation_type = OperationType.INSERT

    def _create_executor(self, line_number, mapping: Entity, values, context, orm):
        # Compile the filtered tree to get the query.
        query = InsertRenderer().render(mapping)

        # is there an extension on the root table?
        if not ReturningClauseRenderer().render(mapping):
            # yield query and split columns
            return SimpleQueryExecutor(line_number, self.operation_type, mapping.name, query, values, context)

        else:

            # create a separate query for the extension
            dependent_query = self._create_extension_insert_query(mapping, values)

            # return dependent query executor to execute both queries
            return DependentQueryExecutor(line_number, self.operation_type, mapping.name, context, (query, values), dependent_query)

    def _create_non_unique_value_dict(self, mapping, row):
        # call super to get all other columns and values
        non_unique_value_dict = super()._create_non_unique_value_dict(mapping, row)

        # remove null and nan values, because we don't need to insert them
        non_unique_value_dict = {key: value for key, value in non_unique_value_dict.items() if not self._is_empty(value)}

        return non_unique_value_dict

    def _create_extension_insert_query(self, mapping, param):
        # get table, name parameter name and values
        table, name_parameter_name, values = ExtensionValueHelper().get_extension_parameter_values(mapping, param)

        # create insert query
        sql = f'insert into {table} (name, module, model, res_id) values (:{name_parameter_name}, :module, :model, :res_id)'
        return sql, values


class UpdateSqlCreator(ExecutorCreator):

    def __init__(self):
        super().__init__()
        self.operation_type = OperationType.UPDATE

    def _create_executor(self, line_number, mapping, values, context, orm):
        # Compile the filtered tree to get the query.
        query = UpdateRenderer().render(mapping)

        # yield query and split columns
        return SimpleQueryExecutor(line_number, self.operation_type, mapping.name, query, values, context)

    def _create_unique_value_dict(self, mapping, row):
        # split row in self and other
        self_row, other_row = self._split_diff_self_other(row)

        # get unique headers
        unique_headers = HeaderRenderer().render_list_unique(mapping)

        # create dictionary with unique column headers as keys and values as values
        self_unique_value_dict = {header: self_row[header] for header in unique_headers}
        other_unique_value_dict = {header: other_row[header] for header in unique_headers}

        # validate that values for self and other are the same, we can't change unique values
        assert self_unique_value_dict == other_unique_value_dict, f'Values for unique columns must be the same for self and other, found: {self_unique_value_dict} and {other_unique_value_dict}'

        return self_unique_value_dict

    def _create_non_unique_value_dict(self, mapping, row):
        # split row in self and other
        self_row, other_row = self._split_diff_self_other(row)

        # get unique columns
        unique_headers = HeaderRenderer().render_list_unique(mapping)

        # get self columns by removing unique headers from self row keys
        non_unique_headers = [header for header in self_row.keys() if header not in unique_headers]

        # create new result dictionary
        non_unique_value_dict = {}

        # iterate non-unique self headers
        for header in non_unique_headers:

            # get self and other values
            self_value = self_row[header]
            other_value = other_row[header]

            # if self and other values are different, add them to the result dictionary
            if self._is_value_modified(self_value, other_value):
                # these values may be <NA>. We need to convert them to None before passing them to the DB, but
                # better to keep them here as <NA> bec/ pandas doesn't like None values. The constructor converts None to <NA>
                # in our test cases.
                non_unique_value_dict[header] = self_value

        # verify that there is at least one modified value
        assert len(non_unique_value_dict) > 0, f'At least one non-unique value must be modified, found: {len(non_unique_value_dict)}'

        return non_unique_value_dict

    def _split_diff_self_other(self, row):
        # extract self and other columns into a new Series
        self_row = row[[column for column in row.index if (column[0] != '__line__' and column[1] in ['', 'self'])]]
        other_row = row[[column for column in row.index if (column[0] != '__line__' and column[1] in ['', 'other'])]]

        # replace header tuples with original column names
        self_row.index = [column[0] for column in self_row.index]
        other_row.index = [column[0] for column in other_row.index]

        # return self and other rows
        return self_row, other_row

    def _is_value_modified(self, this, that):
        # if both values are null, then it's not modified
        if self._is_empty(this) and self._is_empty(that):
            return False

        # if one value is null and the other value is not null, it's modified
        if self._is_empty(this) != self._is_empty(that):
            return True

        # else compare non-null values
        return this != that


class DeleteSqlCreator(ExecutorCreator):
    def __init__(self):
        super().__init__()
        self.operation_type = OperationType.DELETE

    def _create_executor(self, line_number, mapping: Entity, values, context, orm):
        # Compile the filtered tree to get the query.
        query = DeleteRenderer().render(mapping)

        # is there an extension on the root table?
        if not ReturningClauseRenderer().render(mapping):
            # yield query and split columns
            return SimpleQueryExecutor(line_number, self.operation_type, mapping.name, query, values, context)

        else:

            # create a separate query for the extension
            dependent_query = self._create_extension_delete_query(mapping, values)

            # return dependent query executor to execute both queries
            return DependentQueryExecutor(line_number, self.operation_type, mapping.name, context, (query, values), dependent_query)

    def _create_non_unique_value_dict(self, mapping, row):
        # For deletes we don't need non-unique columns.
        # Except, we do need extension name value, because we need to delete the extension even if it's not unique
        return super()._create_non_unique_root_extension_value_dict(mapping, row)

    def _get_line_number(self, row):
        # deletes don't have a matching input row
        return None

    def _create_extension_delete_query(self, mapping, param):
        # get table, name parameter name and values
        table, name_parameter_name, values = ExtensionValueHelper().get_extension_parameter_values(mapping, param)

        # create delete query
        sql = f'delete from {table} where name = :{name_parameter_name} and module = :module and model = :model and res_id = :res_id'
        return sql, values


class ExtensionValueHelper:
    def get_extension_parameter_values(self, mapping: Entity, param):
        # get extension foreign keys
        extension_foreign_keys = [a for a in mapping.attributes if isinstance(a, Reference) and a.extension]

        # only one extension is supported for now
        assert len(extension_foreign_keys) == 1, f'Only one extension is supported, found: {len(extension_foreign_keys)}'
        foreign_key = extension_foreign_keys[0]

        # foreign key must have a single attribute
        assert len(foreign_key.attributes) == 1, f'Foreign key must have a single attribute, found: {len(foreign_key["attributes"])}'
        attribute = foreign_key.attributes[0]
        assert isinstance(attribute, Attribute), f'Foreign key attribute must be an attribute, found: {type(attribute)}'
        name_parameter_name = attribute.parameter

        table = foreign_key.table
        name = param[name_parameter_name]
        qualifier = foreign_key.qualifier
        # TODO: replace with proper way to get the model name
        model = mapping.name.replace('_', '.')

        # create values map
        values = {name_parameter_name: name, 'module': qualifier, 'model': model, 'res_id': None}

        # return table, name parameter name and values
        return table, name_parameter_name, values
